function [U_opt, P_batt] = MPC_optimization(P,V,P_load,SOC)

%% Extrinsic function setting up, running and timing optimization
% Inputs:
%   P       Controller parameters
%   P.simRun  Current simulation run [1,2,3,4]
%   P.omega_B Electrical velocity base [rad/s]
%   P.B       Incidence matrix
%   P.S       Genset rated power [VA]
%   P.M       Genset mechanical startup time [s]
%   P.D       Load frequency damping
%   P.B_sus   Susceptances
%   P.fU      Load power flow selector matrix
%   P.fL      Genset power flow selector matrix
%   P.u_0     Static load setpoint [pu]
%   P.Ts      Control interval [s]
%   P.N       Prediction horizon [steps]
%   P.x_ref   State setpoints
%   P.Q       State deviation weight matrix
%   P.R       Manipulated varable rate weight matrix
%   P.q       Line flow weight
%   V       Controller variables
%   V.x_hat      Estimated state
%   V.P_hat      Predicted load
%   V.U_opt_prev Previous optimal control sequence
% Outputs:
%   U_opt   Optimal control sequence
%   t_calc  Optimization calculation time

% Copyright (c) 2018, Andreas Reason Dahl
% All rights reserved. Published under MIT. See LICENSE.txt.
% Please include the reference given in reference.txt whenever using this.

%% Peak-shaving

% Initialize
P_batt = 0;

% For short load profile
peak_min= 0.1;
peak_max = 0.8;

% For long load profile
%peak_min = 0.15;
%peak_max = 0.75; 

P_min = peak_min*P.total_gen_power;
P_max = peak_max*P.total_gen_power;

SOC_min = 0.2;
SOC_max = 0.8;

% Use battery if within SOC limits
if SOC < SOC_max && P_load < P_min
    % Charge battery
    P_batt =  P_load - P_min;
elseif SOC > SOC_min && P_load > P_max
    % Discharge battery
    P_batt = P_load - P_max;
else % Do not use battery if battery is not within SOC limits    
    P_batt = 0;
end
    
%% Optimize setpoints
obj     = @(u) MPC_objective(u,P,V);

%% Bounds for the gensets
u_min = 0.2; %Unhealthy for the engine to stay below 20%, assume instant change
u_max = 1;

LB      = repmat( u_min*ones(1,P.N), 1, P.N); % Lower bounds 
UB      = repmat( u_max*ones(1,P.N), 1, P.N); % Upper bounds 

options = optimoptions('fmincon','Algorithm','sqp','Display','none','MaxFunEvals', 400000, 'StepTolerance',0.01);

A = repmat([-P.gen1 -P.gen2 -P.gen3 -P.gen4],1,P.N);
b = P.N*(-P_load + P_batt);

U_opt   = fmincon(obj,V.U_opt_prev,A,b,[],[],LB,UB,[],options);

end