function U_opt = MPC_optimization(P,V)

%% Extrinsic function setting up, running and timing optimization
% Inputs:
%   P       Controller parameters
%   P.simRun  Current simulation run [1,2,3,4]
%   P.omega_B Electrical velocity base [rad/s]
%   P.B       Incidence matrix
%   P.S       Genset rated power [VA]
%   P.M       Genset mechanical startup time [s]
%   P.D       Load frequency damping
%   P.B_sus   Susceptances
%   P.fU      Load power flow selector matrix
%   P.fL      Genset power flow selector matrix
%   P.u_0     Static load setpoint [pu]
%   P.Ts      Control interval [s]
%   P.N       Prediction horizon [steps]
%   P.x_ref   State setpoints
%   P.Q       State deviation weight matrix
%   P.R       Manipulated varable rate weight matrix
%   P.q       Line flow weight
%   V       Controller variables
%   V.x_hat      Estimated state
%   V.P_hat      Predicted load
%   V.U_opt_prev Previous optimal control sequence
% Outputs:
%   U_opt   Optimal control sequence
%   t_calc  Optimization calculation time

% Copyright (c) 2018, Andreas Reason Dahl
% All rights reserved. Published under MIT. See LICENSE.txt.
% Please include the reference given in reference.txt whenever using this.

%% Optimize setpoints
obj     = @(u) MPC_objective(u,P,V);

%% Bounds for the gensets
u_min = 0.2; %Unhealthy for the engine to stay below 20%, assume instant change
u_max = 1;

P_MPC_min = -0.05; %plant pu, should be realistic compared to battery size
P_MPC_max = 0.05; %plant pu

LB      = repmat( [u_min*ones(1,4), P_MPC_min], 1, P.N); % Lower bounds
UB      = repmat( [u_max*ones(1,4), P_MPC_max], 1, P.N); % Upper bounds

options = optimoptions('fmincon','Algorithm','sqp','Display','none','MaxFunEvals', 400000, 'StepTolerance',0.01);

A = repmat([-P.gen1 -P.gen2 -P.gen3 -P.gen4 -P.total_gen_power],1,P.N);
b = P.N*(-V.P_load);

initial_value = V.U_opt_prev;

U_opt   = fmincon(obj,initial_value,A,b,[],[],LB,UB,[],options);

end