% Part of this script is taken from Dahl (2018), as stated below, and
% modified by the author, Olav Fiksdahl, in 2020.

% RUNME MPC of Marine Vessel Power System by Use of SPM
%   This script is complimentary to the following conference paper:
%       Dahl, A.R., Thorat, L., and Skjetne, R., 2018. Model Predictive
%       Control of Marine Vessel Power System by Use of Structure
%       Preserving Model. In: Proceedings of the 11th IFAC Control
%       Applications in Marine Systems, Robotics, and Vehicles (CAMS).
%       Opatija, Croatia.
%   Abstract:
%       The paper presents model predictive control (MPC) based on an
%       optimization model drawn from the structure preserving model (SPM)
%       for marine vessel power systems. Three objective functions are
%       proposed to control frequency, transient load and power flow. The
%       objective functions, in addition to a benchmark controller, are
%       applied to a simulated power system.
%   The script sets parameters, runs simulations, determines optimization
%   times, generates plots and tables.

% Copyright (c) 2018, Andreas Reason Dahl
% All rights reserved. Published under MIT. See LICENSE.txt.
% Please include the reference given in reference.txt whenever using this.

close all, clear all, Simulink.sdi.clear, clc
   
%% Plant parameters

f_R = 60; % Rated electrical frequency [Hz]

%% Network

% Initialize network with all gensets connected
y1 = 1; % gensets
y2 = 1;
y3 = 1;
y4 = 1;

b = 1; % bus-tie

% Lines   1   2   3   4   5   6   7     | Nodes | Description
P.B =   [-1   0   0  -1   0   0   0 ;   % 1     | Load 1
          1  -b   0   0 -y2   0   0 ;   % 2     | Load 2
          0   b  -1   0   0 -y3   0 ;   % 3     | Load 3
          0   0   1   0   0   0 -y4 ;   % 4     | Load 4
          0   0   0  y1   0   0   0 ;   % 5     | Genset 1 (diesel)
          0   0   0   0  y2   0   0 ;   % 6     | Genset 2 (gas)
          0   0   0   0   0  y3   0 ;   % 7     | Genset 3 (diesel)
          0   0   0   0   0   0  y4 ;]; % 8     | Genset 4 (gas)

% Initialize, so it can be used later for disconnection
P.y_vec = ones(1,4)';      
      
% Diesel gensets
P.AVR = 1;% can change to make more stable

P.gen1 = 4.9e6; % VA
P.gen2 = 5e6;   % VA
P.gen3 = 5.1e6; % VA
P.gen4 = 5.2e6; % VA

gensets = [ P.gen1  P.gen2   P.gen3    P.gen4 ;   % Rated power [VA]
             3.525   3.525    3.525    3.525 ;];   % Inertia constant [s*W/VA]

P.gensets = gensets;
P.total_gen_power = sum(gensets(1,:));
total_gen_power = P.total_gen_power;

%% MPC
P.Ts       = 1;   % Control interval (timestep) [s] 
P.N        = 4;   % Prediction horizon [steps]

% Setpoint
P.x_ref    = [0   % alpha_1
              0   % alpha_2
              0   % alpha_3
              0   % alpha_4
              0   % alpha_5
              0   % alpha_6
              0   % alpha_7
              0   % alpha_8
              1   % omega_1 Genset 1 frequency [pu]
              1   % omega_2 Genset 2 frequency [pu]
              1   % omega_3 Genset 3 frequency [pu]
              1]; % omega_4 Genset 4 frequency [pu]

% Weights -> High Q = keeps reference freq. High R = small sp. variation
P.Q        = diag([0,0,0,0,0,0,0,0,10,10,10,10]); % q_9, q_10, q_11, q_12

P.R        = zeros(4,4,2);          % r_1, r_2
P.R(:,:,2) = diag([0.1 0.1 0.1 0.1]);       % r_1, r_2 simRun 2

% Optimization model (SPM)
P.omega_B  = 2*pi*f_R;                 % Electrical velocity base [rad/s]
n          = size(P.B,1);              % Number of nodes
m          = size(gensets,2);          %    -"-    generators
n_0        = n-m;                      %    -"-    loads
l          = size(P.B,2);              %    -"-    lines
P.fU       = [eye(n_0) zeros(n_0,m)];  % Load power flow selector matrix
P.fL       = [zeros(m,n_0) eye(m)];    % Genset power flow selector matrix
P.B_sus    = [1/0.0188*ones(l-m,1)     % Line susceptances
              gensets(1,:)'./sum(gensets(1,:)')];
P.D        = diag(ones(n_0,1)*0.0001); % Damping matrix
P.M        = diag(2*gensets(2,:));     % Mechanical starting time (2*H) [s]
P.S        = gensets(1,:)';            % Rated power [VA]

%% Static controller for baseline case
P.u_0      = 0.8*ones(m,1); % Load setpoint [pu]

%% Battery and converter parameters
initial_SOC = 0.5; 
charge_eff = 0.9;
discharge_eff = 0.9;
battery_capacity = 1400; %Ah - high capacity = slow SOC change
battery_voltage = 690; %V
transf_eff = 0.99;
conv_eff = 0.98;

P.k_SOC = 1/(3600*battery_capacity*battery_voltage);
%% Simulation

% Change simulation time here:

short_simulation = true; % Simulate short load profile
%short_simulation = false; % Simulate realistic load profile

if short_simulation == true
    t_simStop = 1000; % [s]
else
    t_simStop = 24*60*60; % [s], 24 hour simulation
end

for simRun = 2 %1:2
    
    if simRun == 1
        disp(['Simulating run ' num2str(simRun) ' out of 2: Baseline case. Control method: Droop control.'])
    elseif simRun == 2
        disp('Simulating with MPC optimization.')
    end
    
    P.simRun = simRun;
    t_simulation(simRun)= tic;
    sim('simscape_model_4gensets_4loads')
    t_simulation(simRun)= toc(t_simulation(simRun));
end

%% Plotting

set(0,'defaulttextinterpreter','latex');

if short_simulation == true
    
% SOC
figure(13), hold on
plot(SOC.Time, 100*SOC.Data(:,1),'LineWidth',1,'Color',uint8([225,220,102]));
legend('Battery SOC','Interpreter','latex')
xlim([0 t_simStop]), ylim([0 100])
xlabel('Time [s]'), 
ylabel('SOC [\%]')

% Generator and SPM frequencies
figure(14), hold on
plot(G1_freq.Time,G1_freq.Data(:,1),'LineWidth',1,'Color',uint8([153,204,255]))
plot(G2_freq.Time,G2_freq.Data(:,1),'LineWidth',1,'Color',uint8([51,153,255]))
plot(G3_freq.Time,G3_freq.Data(:,1),'LineWidth',1,'Color',uint8([0,102,204]))
plot(G4_freq.Time,G4_freq.Data(:,1),'LineWidth',1,'Color',uint8([0,51,102]))
plot(w_SPM_1.Time,w_SPM_1.Data(:,1),'LineWidth',1,'Color',uint8([255,153,153]))
plot(w_SPM_2.Time,w_SPM_2.Data(:,1),'LineWidth',1,'Color',uint8([255,51,51]))
plot(w_SPM_3.Time,w_SPM_3.Data(:,1),'LineWidth',1,'Color',uint8([204,0,0]))
plot(w_SPM_4.Time,w_SPM_4.Data(:,1),'LineWidth',1,'Color',uint8([132,0,0]))
legend('Generator frequency 1','Generator frequency 2','Generator frequency 3','Generator frequency 4',...
    'SPM frequency 1','SPM frequency 2','SPM frequency 3','SPM frequency 4','Interpreter','latex')
xlim([0 t_simStop]), ylim([57 64])
xlabel('Time [s]'), 
ylabel('Electrical frequency [Hz]')

% Load profile
figure(15), hold on
plot(Load_profile.Time, Load_profile.Data(:,1)/1e6,'LineWidth',1,'Color',uint8([0,204,102]));
legend('Load profile','Interpreter','latex')
xlim([0 t_simStop]), ylim([0 total_gen_power/1e6])
xlabel('Time [s]'), 
ylabel('Load [MW]')

% Generator setpoints
figure(16), hold on
stairs(u1.Time,u1.Data(:,1),'LineWidth',1,'Color',uint8([255,153,153]))
stairs(u2.Time,u2.Data(:,1),'LineWidth',1,'Color',uint8([255,51,51]))
stairs(u3.Time,u3.Data(:,1),'LineWidth',1,'Color',uint8([204,0,0]))
stairs(u4.Time,u4.Data(:,1),'LineWidth',1,'Color',uint8([132,0,0]))
legend('Generator 1','Generator 2','Generator 3','Generator 4','Interpreter','latex')
xlim([0 t_simStop]), ylim([0 1])
xlabel('Time [s]'), 
ylabel('Load setpoint [pu]')

% Power flow
figure(17), hold on
stairs(Power_G1.Time,Power_G1.Data(:,1),'LineWidth',1,'Color',uint8([153,204,255]))
stairs(Power_G2.Time,Power_G2.Data(:,1),'LineWidth',1,'Color',uint8([51,153,255]))
stairs(Power_G3.Time,Power_G3.Data(:,1),'LineWidth',1,'Color',uint8([0,102,204]))
stairs(Power_G4.Time,Power_G4.Data(:,1),'LineWidth',1,'Color',uint8([0,51,102]))
legend('Generator 1','Generator 2','Generator 3','Generator 4','Interpreter','latex')
xlim([0 t_simStop]), ylim([0 5.2e6])
xlabel('Time [s]'), 
ylabel('Power flow [MW]')

% Battery power flow
figure(18), hold on
plot(P_batt.Time, P_batt.Data(:,1),'LineWidth',1,'Color',uint8([225,220,102]));
legend('Battery power','Interpreter','latex')
xlim([0 t_simStop]), ylim([-2.2e6 2.2e6])
xlabel('Time [s]'), 
ylabel('Battery power flow [MW]')

else
    
% SOC
figure(13), hold on
plot(SOC.Time/3600, 100*SOC.Data(:,1),'LineWidth',1,'Color',uint8([225,220,102]));
legend('Battery SOC','Interpreter','latex')
xlim([0 t_simStop/3600]), ylim([0 100])
xlabel('Time [h]'), 
ylabel('SOC [\%]')

% Generator and SPM frequencies
figure(14), hold on
plot(G1_freq.Time/3600,G1_freq.Data(:,1),'LineWidth',1,'Color',uint8([153,204,255]))
plot(G2_freq.Time/3600,G2_freq.Data(:,1),'LineWidth',1,'Color',uint8([51,153,255]))
plot(G3_freq.Time/3600,G3_freq.Data(:,1),'LineWidth',1,'Color',uint8([0,102,204]))
plot(G4_freq.Time/3600,G4_freq.Data(:,1),'LineWidth',1,'Color',uint8([0,51,102]))
plot(w_SPM_1.Time/3600,w_SPM_1.Data(:,1),'LineWidth',1,'Color',uint8([255,153,153]))
plot(w_SPM_2.Time/3600,w_SPM_2.Data(:,1),'LineWidth',1,'Color',uint8([255,51,51]))
plot(w_SPM_3.Time/3600,w_SPM_3.Data(:,1),'LineWidth',1,'Color',uint8([204,0,0]))
plot(w_SPM_4.Time/3600,w_SPM_4.Data(:,1),'LineWidth',1,'Color',uint8([132,0,0]))
legend('Generator frequency 1','Generator frequency 2','Generator frequency 3','Generator frequency 4',...
    'SPM frequency 1','SPM frequency 2','SPM frequency 3','SPM frequency 4','Interpreter','latex')
xlim([0 t_simStop/3600]), ylim([57 64])
xlabel('Time [h]'), 
ylabel('Electrical frequency [Hz]')

% Load profile
figure(15), hold on
plot(Load_profile.Time/3600, Load_profile.Data(:,1)/1e6,'LineWidth',1,'Color',uint8([0,204,102]));
legend('Load profile','Interpreter','latex')
xlim([0 t_simStop/3600]), ylim([0 total_gen_power/1e6])
xlabel('Time [h]'), 
ylabel('Load [MW]')

% Generator setpoints
figure(16), hold on
stairs(u1.Time/3600,u1.Data(:,1),'LineWidth',1,'Color',uint8([255,153,153]))
stairs(u2.Time/3600,u2.Data(:,1),'LineWidth',1,'Color',uint8([255,51,51]))
stairs(u3.Time/3600,u3.Data(:,1),'LineWidth',1,'Color',uint8([204,0,0]))
stairs(u4.Time/3600,u4.Data(:,1),'LineWidth',1,'Color',uint8([132,0,0]))
legend('Generator 1','Generator 2','Generator 3','Generator 4','Interpreter','latex')
xlim([0 t_simStop/3600]), ylim([0 1])
xlabel('Time [h]'), 
ylabel('Load setpoint [pu]')

% Power flow
figure(17), hold on
stairs(Power_G1.Time/3600,Power_G1.Data(:,1),'LineWidth',1,'Color',uint8([153,204,255]))
stairs(Power_G2.Time/3600,Power_G2.Data(:,1),'LineWidth',1,'Color',uint8([51,153,255]))
stairs(Power_G3.Time/3600,Power_G3.Data(:,1),'LineWidth',1,'Color',uint8([0,102,204]))
stairs(Power_G4.Time/3600,Power_G4.Data(:,1),'LineWidth',1,'Color',uint8([0,51,102]))
legend('Generator 1','Generator 2','Generator 3','Generator 4','Interpreter','latex')
xlim([0 t_simStop/3600]), ylim([0 5.2e6])
xlabel('Time [h]'), 
ylabel('Power flow [MW]')

% Battery power flow
figure(18), hold on
plot(P_batt.Time/3600, P_batt.Data(:,1),'LineWidth',1,'Color',uint8([225,220,102]));
legend('Battery power','Interpreter','latex')
xlim([0 t_simStop/3600]), ylim([-2.2e6 2.2e6])
xlabel('Time [h]'), 
ylabel('Battery power flow [MW]')

end % if plot